package ru.dbotthepony.kstarbound.world

import com.google.common.base.Predicate
import it.unimi.dsi.fastutil.longs.Long2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet
import it.unimi.dsi.fastutil.objects.ReferenceOpenHashSet
import ru.dbotthepony.kbox2d.api.ContactImpulse
import ru.dbotthepony.kbox2d.api.IContactFilter
import ru.dbotthepony.kbox2d.api.IContactListener
import ru.dbotthepony.kbox2d.api.Manifold
import ru.dbotthepony.kbox2d.dynamics.B2Fixture
import ru.dbotthepony.kbox2d.dynamics.B2World
import ru.dbotthepony.kbox2d.dynamics.contact.AbstractContact
import ru.dbotthepony.kstarbound.math.*
import ru.dbotthepony.kstarbound.util.Timer
import ru.dbotthepony.kstarbound.world.api.CHUNK_SIZE
import ru.dbotthepony.kstarbound.world.api.CHUNK_SIZE_BITS
import ru.dbotthepony.kstarbound.world.api.CHUNK_SIZE_MASK
import ru.dbotthepony.kstarbound.world.api.ICellAccess
import ru.dbotthepony.kstarbound.world.api.IChunkCell
import ru.dbotthepony.kstarbound.world.api.TileView
import ru.dbotthepony.kstarbound.world.entities.Entity
import ru.dbotthepony.kvector.api.IStruct2d
import ru.dbotthepony.kvector.api.IStruct2i
import ru.dbotthepony.kvector.arrays.Int2DArray
import ru.dbotthepony.kvector.arrays.Object2DArray
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.util2d.AABBi
import ru.dbotthepony.kvector.vector.Vector2d
import ru.dbotthepony.kvector.vector.Vector2i

abstract class World<This : World<This, ChunkType>, ChunkType : Chunk<This, ChunkType>>(
	val seed: Long,
	val size: Vector2i?,
	val loopX: Boolean,
	val loopY: Boolean
) {
	abstract class AbstractCoordinatesWrapper {
		abstract fun cell(value: Int): Int
		abstract fun cell(value: Double): Double
		abstract fun cell(value: Float): Float

		abstract fun chunk(value: Int): Int
		abstract fun cellOffset(chunk: Int): Int

		abstract fun inBounds(value: Int): Boolean
		abstract fun inBoundsChunk(value: Int): Boolean

		protected fun positiveModulo(a: Int, b: Int): Int {
			val result = a % b
			return if (result < 0) result + b else result
		}

		protected fun positiveModulo(a: Double, b: Int): Double {
			val result = a % b
			return if (result < 0.0) result + b else result
		}

		protected fun positiveModulo(a: Float, b: Int): Float {
			val result = a % b
			return if (result < 0f) result + b else result
		}
	}

	object PassthroughWrapper : AbstractCoordinatesWrapper() {
		override fun cell(value: Int): Int = value
		override fun cell(value: Double): Double = value
		override fun cell(value: Float): Float = value
		override fun chunk(value: Int): Int = value
		override fun cellOffset(chunk: Int): Int = 0

		override fun inBounds(value: Int) = true
		override fun inBoundsChunk(value: Int) = true
	}

	class CoordinatesWrapper(private val cell: Int, private val chunk: Int) : AbstractCoordinatesWrapper() {
		override fun inBounds(value: Int): Boolean {
			return value in 0 until cell
		}

		override fun inBoundsChunk(value: Int): Boolean {
			return value in 0 until chunk
		}

		private val misalignedTiles = cell and CHUNK_SIZE_MASK
		private val misalignedTiles2 = misalignedTiles - CHUNK_SIZE

		override fun cellOffset(chunk: Int): Int {
			if (misalignedTiles == 0) {
				return 0
			} else if (chunk >= 0) {
				return (chunk / this.chunk) * misalignedTiles2
			} else {
				return ((chunk + 1) / this.chunk - 1) * misalignedTiles2
			}
		}

		override fun cell(value: Int): Int {
			return positiveModulo(value, cell)
		}

		override fun cell(value: Double): Double {
			return positiveModulo(value, cell)
		}

		override fun cell(value: Float): Float {
			return positiveModulo(value, cell)
		}

		override fun chunk(value: Int): Int {
			return positiveModulo(value, chunk)
		}
	}

	class CoordinatesClamper(private val cell: Int, private val chunk: Int) : AbstractCoordinatesWrapper() {
		override fun inBounds(value: Int): Boolean {
			return value in 0 until cell
		}

		override fun cellOffset(chunk: Int): Int {
			return 0
		}

		override fun inBoundsChunk(value: Int): Boolean {
			return value in 0 until chunk
		}

		override fun cell(value: Int): Int {
			return value.coerceIn(0, cell - 1)
		}

		override fun cell(value: Double): Double {
			return value.coerceIn(0.0, cell - 1.0)
		}

		override fun cell(value: Float): Float {
			return value.coerceIn(0f, cell - 1f)
		}

		override fun chunk(value: Int): Int {
			return value.coerceIn(0, chunk - 1)
		}
	}

	abstract inner class ChunkMap : ICellAccess {
		abstract val x: AbstractCoordinatesWrapper
		abstract val y: AbstractCoordinatesWrapper

		fun inBounds(x: Int, y: Int) = this.x.inBounds(x) && this.y.inBounds(y)
		fun inBounds(value: IStruct2i) = this.x.inBounds(value.component1()) && this.y.inBounds(value.component2())

		override fun randomLongFor(x: Int, y: Int): Long {
			return super.randomLongFor(x, y) xor seed
		}

		val background = TileView.Background(this)
		val foreground = TileView.Foreground(this)

		abstract operator fun get(x: Int, y: Int): ChunkType?
		operator fun get(pos: ChunkPos) = get(pos.x, pos.y)

		open fun chunkFromCell(x: Int, y: Int): ChunkType? {
			return get(ChunkPos.component(x), ChunkPos.component(y))
		}

		fun chunkFromCell(pos: IStruct2i) = chunkFromCell(pos.component1(), pos.component2())

		override fun getCell(x: Int, y: Int): IChunkCell {
			return chunkFromCell(x, y)?.getCell(x and CHUNK_SIZE_MASK, y and CHUNK_SIZE_MASK) ?: IChunkCell.Companion
		}

		abstract operator fun set(x: Int, y: Int, chunk: ChunkType)
		abstract fun remove(x: Int, y: Int)

		fun cellToGrid(x: Int, y: Int) = ChunkPos.fromPosition(this.x.cell(x), this.y.cell(y))
		fun cellToGrid(x: Double, y: Double) = ChunkPos.fromPosition(this.x.cell(x), this.y.cell(y))
		fun cellToGrid(position: IStruct2i) = cellToGrid(position.component1(), position.component2())
		fun cellToGrid(position: IStruct2d) = cellToGrid(position.component1(), position.component2())

		fun computeIfAbsent(pos: ChunkPos) = computeIfAbsent(pos.x, pos.y)

		fun computeIfAbsent(x: Int, y: Int): ChunkType {
			val existing = get(x, y)

			if (existing != null)
				return existing

			val pos = ChunkPos(this.x.chunk(x), this.y.chunk(y))
			val chunk = chunkFactory(pos)
			val orphanedInThisChunk = ArrayList<Entity>()

			for (ent in orphanedEntities) {
				val (ex, ey) = ent.position

				if (ChunkPos.fromPosition(this.x.cell(ex), this.y.cell(ey)) == pos) {
					orphanedInThisChunk.add(ent)
				}
			}

			for (ent in orphanedInThisChunk) {
				ent.chunk = chunk
			}

			set(x, y, chunk)
			return chunk
		}
	}

	// infinite chunk map is around 30% slower than rectangular one
	inner class InfiniteChunkMap : ChunkMap() {
		private val map = Long2ObjectOpenHashMap<ChunkType>()

		override fun get(x: Int, y: Int): ChunkType? {
			return map[ChunkPos.toLong(x, y)]
		}

		override fun set(x: Int, y: Int, chunk: ChunkType) {
			map[ChunkPos.toLong(x, y)] = chunk
		}

		override fun remove(x: Int, y: Int) {
			map.remove(ChunkPos.toLong(x, y))
		}

		override val x = PassthroughWrapper
		override val y = PassthroughWrapper
	}

	inner class RectChunkMap : ChunkMap() {
		val width = size!!.x
		val height = size!!.y

		val widthInChunks = if (width and CHUNK_SIZE_MASK == 0) width / CHUNK_SIZE else width / CHUNK_SIZE + 1
		val heightInChunks = if (height and CHUNK_SIZE_MASK == 0) height / CHUNK_SIZE else height / CHUNK_SIZE + 1

		override val x: AbstractCoordinatesWrapper = if (loopX) CoordinatesWrapper(width, widthInChunks) else CoordinatesClamper(width, widthInChunks)
		override val y: AbstractCoordinatesWrapper = if (loopY) CoordinatesWrapper(height, heightInChunks) else CoordinatesClamper(height, heightInChunks)

		private val map = Object2DArray.nulls<ChunkType>(widthInChunks, heightInChunks)

		override fun get(x: Int, y: Int): ChunkType? {
			return map[this.x.chunk(x), this.y.chunk(y)]
		}

		override fun chunkFromCell(x: Int, y: Int): ChunkType? {
			val ix = this.x.cell(x)
			val iy = this.y.cell(y)
			return get(ix ushr CHUNK_SIZE_BITS, iy ushr CHUNK_SIZE_BITS)
		}

		override fun getCell(x: Int, y: Int): IChunkCell {
			val ix = this.x.cell(x)
			val iy = this.y.cell(y)
			return get(ix ushr CHUNK_SIZE_BITS, iy ushr CHUNK_SIZE_BITS)?.getCell(ix and CHUNK_SIZE_MASK, iy and CHUNK_SIZE_MASK) ?: IChunkCell.Companion
		}

		override fun set(x: Int, y: Int, chunk: ChunkType) {
			map[this.x.chunk(x), this.y.chunk(y)] = chunk
		}

		override fun remove(x: Int, y: Int) {
			map[this.x.chunk(x), this.y.chunk(y)] = null
		}
	}

	val chunkMap: ChunkMap = if (size != null) RectChunkMap() else InfiniteChunkMap()

	/**
	 * Chunks, which have their collision mesh changed
	 */
	val dirtyPhysicsChunks = ObjectOpenHashSet<ChunkType>()

	val physics = B2World(Vector2d(0.0, -EARTH_FREEFALL_ACCELERATION))

	private var timers = ArrayList<Timer>()

	fun timer(period: Double, executionTimes: Int, func: (Timer) -> Unit): Timer {
		val timer = Timer(period, executionTimes, func)
		timers.add(timer)
		return timer
	}

	init {
		physics.contactFilter = object : IContactFilter {
			override fun shouldCollide(fixtureA: B2Fixture, fixtureB: B2Fixture): Boolean {
				val dataA = fixtureA.body!!.userData
				val dataB = fixtureB.body!!.userData

				//if (dataA is AbstractProjectileMovementController && dataB is AbstractProjectileMovementController) {
				//	return false
				//}

				return true
			}
		}

		physics.contactListener = object : IContactListener {
			override fun beginContact(contact: AbstractContact) {
				val dataA = contact.fixtureA.body!!.userData
				val dataB = contact.fixtureB.body!!.userData

				if (dataA is IContactListener) {
					dataA.beginContact(contact)
				}

				if (dataB is IContactListener) {
					dataB.beginContact(contact)
				}
			}

			override fun endContact(contact: AbstractContact) {
				val dataA = contact.fixtureA.body!!.userData
				val dataB = contact.fixtureB.body!!.userData

				if (dataA is IContactListener) {
					dataA.endContact(contact)
				}

				if (dataB is IContactListener) {
					dataB.endContact(contact)
				}
			}

			override fun preSolve(contact: AbstractContact, oldManifold: Manifold) {
				val dataA = contact.fixtureA.body!!.userData
				val dataB = contact.fixtureB.body!!.userData

				if (dataA is IContactListener) {
					dataA.preSolve(contact, oldManifold)
				}

				if (dataB is IContactListener) {
					dataB.preSolve(contact, oldManifold)
				}
			}

			override fun postSolve(contact: AbstractContact, impulse: ContactImpulse) {
				val dataA = contact.fixtureA.body!!.userData
				val dataB = contact.fixtureB.body!!.userData

				if (dataA is IContactListener) {
					dataA.postSolve(contact, impulse)
				}

				if (dataB is IContactListener) {
					dataB.postSolve(contact, impulse)
				}
			}
		}
	}

	/**
	 * Таймер этого мира, в секундах.
	 *
	 * Условия:
	 * * Это значение МОЖЕТ уменьшаться, в том числе в силу того, что клиент
	 * обсчитывает мир быстрее, чем сервер, и наоборот.
	 * * Точка отсчёта не имеет смысла (по умолчанию отсчитывается от нуля).
	 * * Все сущности обязаны использовать этот таймер для определения времени своих событий
	 * * Таймер НЕ МОЖЕТ возвращать разные значения в пределах одного тика.
	 * т.е. таймер должен быть обновлён только после или перед тиком всех потомков
	 * * НЕ ОБЯЗАН соответствовать течению реального времени, но как правило, соответствует ему
	 * (если позволяет скорость обсчёта потомков).
	 * * ОБЯЗАН быть сохранён при сохранении мира, и выставляться при его загрузке.
	 */
	var timer: Double = 0.0
		protected set

	/**
	 * Заставляет этот мир "думать" на заданное количество секунд.
	 * Вызывает все необходимые методы у потомков, обновляя игровое состояние.
	 */
	fun think(delta: Double) {
		require(delta > 0.0) { "Tried to update $this by $delta seconds" }

		try {
			for (chunk in dirtyPhysicsChunks) {
				chunk.bakeCollisions()
			}

			dirtyPhysicsChunks.clear()

			physics.step(delta, 6, 4)

			timer += delta

			if (timers.isNotEmpty()) {
				val current = timers
				timers = ArrayList()

				val iterator = current.iterator()

				for (timer in iterator) {
					timer.think(delta)

					if (timer.destroyed) {
						iterator.remove()
					}
				}

				if (timers.isNotEmpty()) {
					current.addAll(timers)
				}

				timers = current
			}

			thinkInner(delta)
		} catch(err: Throwable) {
			throw RuntimeException("Ticking world $this", err)
		}
	}

	protected abstract fun thinkInner(delta: Double)

	/**
	 * Сущности, у которых нет чанка, но они находятся в этом мире
	 */
	val orphanedEntities = ReferenceOpenHashSet<Entity>()

	/**
	 * Сущности, находящиеся в этом мире
	 */
	val entities = ReferenceOpenHashSet<Entity>()

	/**
	 * Стандартное ускорение свободного падения в Starbound Units/секунда^2
	 *
	 * При Vector2d.ZERO = невесомость
	 */
	var gravity by physics::gravity

	protected abstract fun chunkFactory(pos: ChunkPos): ChunkType

	/**
	 * Возвращает все чанки, которые пересекаются с заданным [boundingBox]
	 */
	fun collectWithPosition(boundingBox: AABBi): List<Pair<ChunkPos, ChunkType>> {
		val output = ArrayList<Pair<ChunkPos, ChunkType>>()

		for (pos in boundingBox.chunkPositions) {
			val chunk = chunkMap[pos]

			if (chunk != null && (loopX || chunkMap.x.inBoundsChunk(pos.x)) && (loopY || chunkMap.y.inBoundsChunk(pos.y))) {
				output.add(pos to chunk)
			}
		}

		return output
	}

	fun testSpace(aabb: AABB, predicate: Predicate<IChunkCell> = Predicate { it.foreground.material != null }): Boolean {
		val tiles = aabb.encasingIntAABB()

		for (x in tiles.mins.x .. tiles.maxs.x) {
			for (y in tiles.mins.y .. tiles.maxs.y) {
				if (predicate.test(chunkMap.getCell(x, y))) {
					return true
				}
			}
		}

		return false
	}

	// Свет

	private fun floodLightInto(
		lightmap: Int2DArray,
		thisIntensity: Int,
		lightBlockerStrength: Int,
		posX: Int,
		worldPosX: Int,
		posY: Int,
		worldPosY: Int,
	): Int {
		if (lightmap[posX, posY] >= thisIntensity) {
			return 1
		}

		val tile = chunkMap.getCell(worldPosX, worldPosY)

		var newIntensity: Int

		if (tile.foreground.material?.renderParameters?.lightTransparent == false) {
			newIntensity = thisIntensity - lightBlockerStrength
		} else {
			newIntensity = thisIntensity - 1
		}

		if (tile.foreground.material != null)
			newIntensity = 0

		lightmap[posX, posY] = newIntensity.coerceAtLeast(0)

		if (newIntensity > 1) {
			var c = 1

			c += floodLightInto(
				lightmap, newIntensity, lightBlockerStrength,
				posX + 1,
				worldPosX + 1,
				posY,
				worldPosY,
			)

			c += floodLightInto(
				lightmap, newIntensity, lightBlockerStrength,
				posX - 1,
				worldPosX - 1,
				posY,
				worldPosY,
			)

			c += floodLightInto(
				lightmap, newIntensity, lightBlockerStrength,
				posX,
				worldPosX,
				posY + 1,
				worldPosY + 1,
			)

			c += floodLightInto(
				lightmap, newIntensity, lightBlockerStrength,
				posX,
				worldPosX,
				posY - 1,
				worldPosY - 1,
			)

			return c
		}

		return 1
	}

	/**
	 * Просчитывает распространение света во все стороны на указанной позиции (в тайлах)
	 *
	 * [lightIntensity] - максимальное расстояние, которое может пройти свет из точки своего появления.
	 * Имеет жёсткое ограничение в [CHUNK_SIZE].
	 *
	 * [lightBlockerStrength] - какова стоимость "пробития" тайла насквозь, который не пропускает свет
	 */
	fun floodLight(
		lightPosition: Vector2i,
		lightIntensity: Int,
		lightBlockerStrength: Int = 4,
	): Int2DArray {
		require(lightIntensity >= 1) { "Invalid light intensity $lightIntensity" }
		require(lightBlockerStrength >= 1) { "Invalid light blocker strength $lightBlockerStrength" }
		require(lightIntensity <= CHUNK_SIZE) { "Too intensive light! $lightIntensity" }

		val lightmap = Int2DArray.allocate(lightIntensity * 2 + 1, lightIntensity * 2 + 1)

		floodLightInto(
			lightmap,
			lightIntensity,
			lightBlockerStrength,
			lightIntensity,
			lightPosition.x,
			lightIntensity,
			lightPosition.y,
		)

		return lightmap
	}
}
